@@ -2,14 +2,14 @@ package Parser::FIT;
 
 use strict;
 use warnings;
-use Carp qw/croak carp/;
+use Carp qw/croak confess/;
 use feature 'state';
 use Math::BigInt;
 use Parser::FIT::Profile;
 
 #require "Profile.pm";
 
-our $VERSION = 0.09;
+our $VERSION = 0.10;
 
 sub new {
 	my $class = shift;
@@ -274,7 +274,7 @@ sub _parse_definition_message {
 
 	$self->_debug("DefinitionMessageHeader:");
 	$self->_debug("Arch: $arch - GlobalMessage: " . (defined $globalMessageType ? $globalMessageType->{name} : "<UNKNOWN_GLOBAL_MESSAGE>") . " ($globalMessageId) - #Fields: $fields");
-	carp "BigEndian isn't supported so far!" if($arch == 1);
+	confess "BigEndian isn't supported so far!" if($arch == 1);
 
 	my ($messageFields, $devMsgFields, $recordLength) = ([], [], 0);
 	my @fields;
@@ -289,11 +289,30 @@ sub _parse_definition_message {
 
 	my $combinedDataFields = [@$messageFields, @$devMsgFields];
 
+	my $buildPackTemplate = sub {
+		my $dataField = shift;
+
+		my $templateChar = $dataField->{baseType}->{packTemplate};
+		my $arrayLength = $dataField->{arrayLength};
+		my $fieldName = $dataField->{fieldDescriptor}->{name};
+		my $fieldDatatype = $dataField->{fieldDescriptor}->{type} || "<UNKNOWN_DATA_TYPE>";
+
+		my $templateString = "${templateChar}";
+
+		if($arrayLength > 1) {
+			$templateString .= "[${arrayLength}]";
+		}
+
+		$templateString .= " # $fieldName baseType=$fieldDatatype baseTypeSize=$dataField->{baseType}->{size}, storageSize=$dataField->{storageSize}, byteOffsetInMessage=" . $dataField->{byteOffsetInMessage};
+
+		return $templateString;
+	};
+
 	my $localMessage = {
 		size => $recordLength,
 		dataFields => $combinedDataFields,
 		globalMessage => $globalMessageType,,
-		unpackTemplate => join("", map { $_->{baseType}->{packTemplate} . '[' . $_->{arrayLength} . ']' } @$combinedDataFields),
+		unpackTemplate => join("\n", map { $buildPackTemplate->($_) } @$combinedDataFields),
 		isDeveloperMessage => $header->{isDeveloperData},
 		isUnknownMessage => !defined $globalMessageType,
 	};
@@ -329,27 +348,39 @@ sub _parse_defintion_message_fields {
 
 	my @dataFields;
 
-	foreach(1..$numberOfFields) {
+	my $byteOffsetInMessage = 0;
+	foreach my $i (1..$numberOfFields) {
 		my $fieldDefinitionData = $self->_readBytes(3); # Every Field has 3 Bytes
-		my ($fieldDefinition, $size, $baseTypeData)  = unpack("CCc", $fieldDefinitionData);
+		my ($fieldDefinitionId, $size, $baseTypeData)  = unpack("CCc", $fieldDefinitionData);
 		my ($baseTypeEndian, $baseTypeNumber) = ($baseTypeData & 128, $baseTypeData & 15);
 		my $baseType = $self->_get_base_type($baseTypeNumber);
-		my $fieldDescriptor = $fieldDefinitions->{$fieldDefinition};
+		my $fieldDescriptor = $fieldDefinitions->{$fieldDefinitionId};
 
 		die "Failed to parse file: Size=$size cannot be 0 or less" if($size <= 0);
 
 		if(!defined $fieldDescriptor) {
 			$fieldDescriptor = {
 				isUnkownField => 1,
-				name => "<UNKNOWN_FIELD_NAME>"
+				name => "<UNKNOWN_FIELD_NAME>",
+				type => $baseType->{name},
 			};
 		}
 
 		my $fieldName = $fieldDescriptor->{name};
-		$self->_debug("FieldDefinition: Nr: $fieldDefinition (" . $fieldName . "), Size: $size, BaseType: " . $baseType->{name} . " ($baseTypeNumber), BaseTypeEndian: $baseTypeEndian");
+		$self->_debug("FieldDefinition[$i]: $fieldName ($fieldDefinitionId), Size: $size, BaseType: " . $baseType->{name} . " ($baseTypeNumber), BaseTypeEndian: $baseTypeEndian");
 		$recordLength += $size;
 
-		push(@dataFields, { baseType => $baseType, storageSize => $size, isArray => $size > $baseType->{size}, arrayLength => $size/$baseType->{size}, fieldDescriptor => $fieldDescriptor });
+		my $fieldDefinition = {
+			baseType => $baseType,
+			storageSize => $size,
+			isArray => $size > $baseType->{size},
+			arrayLength => $size/$baseType->{size},
+			fieldDescriptor => $fieldDescriptor,
+			byteOffsetInMessage => $byteOffsetInMessage
+		};
+
+		push(@dataFields, $fieldDefinition);
+		$byteOffsetInMessage += $size;
 	}
 
 	return (\@dataFields, $recordLength);
@@ -514,9 +545,22 @@ sub _parse_local_message_record {
 	my %result;
 
 	my $fieldCount = scalar @{$localMessage->{dataFields}};
+	my $rawValueIndex = 0;
 	for(my $i = 0; $i < $fieldCount; $i++) {
 		my $localMessageField = $localMessage->{dataFields}->[$i];
-		my $rawValue = $rawFields[$i];
+		my $rawValue = $rawFields[$rawValueIndex];
+
+		# Special case for strings:
+		# The pack/unpack template for strings looks like an array (e.G. Z[32]) but does not produce
+		# multiple entries after unpacking, but only a single one.
+		# Therefore after reading a string we should only proceed to the next entry.
+		# While reading a uint8 array (e.G. c[4]) we must skip 4 entries to arrive at the next field.
+		if($localMessageField->{baseType}->{name} eq "string") {
+			$rawValueIndex += 1;
+		}
+		else {
+			$rawValueIndex += $localMessageField->{arrayLength};
+		}
 
 		my $fieldDescriptor = $localMessageField->{fieldDescriptor};
 		my $fieldName = $fieldDescriptor->{name};
@@ -525,12 +569,20 @@ sub _parse_local_message_record {
 			next;
 		}
 
-		my $postProcessedValue = $self->postProcessRawValue($rawValue, $fieldDescriptor);
+		my $postProcessedValue = undef;
+		my $baseType = $localMessageField->{baseType};
+
+		my $isValid = $self->_check_for_valid_value($baseType, $rawValue);
+
+		if($isValid) {
+			$postProcessedValue = $self->postProcessRawValue($rawValue, $fieldDescriptor);
+		}
 
 		$result{$fieldName} = {
 			value => $postProcessedValue,
 			rawValue => $rawValue,
 			fieldDescriptor => $fieldDescriptor,
+			isInvalid => !$isValid,
 		};
 	}
 
@@ -540,6 +592,20 @@ sub _parse_local_message_record {
 	};
 }
 
+sub _check_for_valid_value {
+	my $self = shift;
+	my $baseType = shift;
+	my $rawValue = shift;
+
+	# All strings are good.
+	if($baseType->{name} eq "string") {
+		return 1;
+	}
+
+	my $isValid = $baseType->{invalid} != $rawValue;
+	return $isValid;
+}
+
 sub postProcessRawValue {
 	my $self = shift;
 	my $rawValue = shift;
@@ -637,6 +703,7 @@ sub _get_base_type {
 			name => "uint8z",
 			size => 1,
 			invalid => 0x00,
+			# TODO this should be a capital C since it is unsigned!
 			packTemplate => "c"
 		},
 		{
